#ifndef	DV_INET_ADDR_H
#define	DV_INET_ADDR_H

// $Id: inetaddress.h,v 1.18 2003/12/10 06:26:12 dvermeir Exp $

#include	<string>
#include	<dvutil/ref.h>
/** @mainpage 

<a href="../download/">Download</a>

@author dvermeir@vub.ac.be
    
@section install Installation

Standard. See the INSTALL file in the top directory.
This package depends on the dvutil package.

@section intro Introduction

This package provides basic network classes.
See the ``Compound Members'' entry in the top menubar.

<ul>
<li> inetaddress.h 
  
  Declares a class @ref Dv::Net::InetAddress that represents an internet
  host. It encapsulates gethostbyname and friends.

<li> socket.h

  Declares an iostream specialization  @ref Dv::Net::Socket that represents a
  network connection. Opening a connection is as simple as
  @code
  Dv::Net::Socket so("server-host",server-port);
  if (so) { 
    so << "send request";
    std::string reply;
    so >> reply;
    }
  else 
    std::cerr << "Connection failed: " << so.strerror() << std::endl;
  @endcode

<li> serversocket.h
  Declares an class \ref Dv::Net::ServerSocket that represents an
  internet server.
  
  @code
  try {
    Dv::Net::ServerSocket server(server-port);
    Dv::Util::ref<Socket> client(server.accept());
    std::cerr << "Connection from " << client->host() << ":" << client->port() << std::endl;

    std::string request;
    std::string reply;

    while (*client>>request)
      *client << reply;
    }
  catch (Dv::Net::ServerSocketError& e) {
    cerr << e.what() << endl;
    return 1;
    }
  @endcode
<li> usocket.h and userversocket.h provide similar classes: Dv::Net::usocket
  and Dv::Net::userversocket for unix sockets.
<li> A simple generic protocol for transfering arbitrary objects over
a connection is also provided in message.h . Any class @a A that
support <code>operator<<(ostream&, const A&)</code>
and <code>operator>>(istream&, A&)</code> can be sent and received over
a socket (or another stream) using code such as the following.
 @code
  A a(..);
  // Send an A object. 
  Dv::Net::Message<A> m_out(a);
  if (! m_out.send(socket) )
    throw std::runtime_error("error sending");
  try {
    // Receive an A object.
    Dv::Net::Message<A> m_in;
    m_in.receive(socket);
    }
  catch (std::runtime_error e) { // error receiving
    std::cerr << e.what() << std::endl;
    }
  A a_in(m_in);
 @endcode
</ul>
*/

/** @file 
 *  The Dv::Net::InetAddress class represents a valid internet host.
 */
namespace Dv {
/**  Dv::Net is the namepace of this package. */
namespace Net {

/** A Dv::Net::InetAddress object represents a valid internet host. */
class InetAddress {
public:
  /**
   * Factory method: return ref<InetAddress> corresponding with host or 0.
   * @param host string representation of host, of the form
   *   "tinf2.vub.ac.be" or "134.184.65.2".  "localhost" is also a valid argument.
   * @return reference to InetAddress representing the host with given
   *   name or 0
   *
   *  Example usage:
   *  @code
   *  Dv::Util::ref<InetAddress>  address(InetAddress::Net::by_name("tinf2.vub.ac.be"));
   *  if (address) 
   *    cout << address->dot_address() << endl;
   *  @endcode
   */
  static Dv::Util::ref<InetAddress> by_name(const std::string& host);
  /** Factory method: return ref<InetAddress> corresponding with addr or 0.
   * @param addr numeric internet address
   * @return reference to InetAddress representing the host with @a addr or 0.
   */
  static Dv::Util::ref<InetAddress> by_address(unsigned long addr);

  /** @return true iff the numeric addresses are the same. */
  bool operator==(const InetAddress&) const;

  /** @return string representation of host, e.g. "tinf2.vub.ac.be." */
  const std::string&  host() const { return host_; }
  /** @return numeric address of host (in network order). */
  unsigned long address() const	{ return address_; }
  /** @return dot address of host, e.g. "134.184.65.2". */
  const std::string& dot_address() const { return dot_address_; }
  /** @return name of local host or "localhost" if not found. */
  static std::string local_host();
private:
  std::string  host_; // star2.vub.ac.be
  unsigned long  address_; // in network order
  std::string  dot_address_; // 134.184.65.2

  /** Private constructor. */
  InetAddress(const char*, unsigned long, const char *);
};

}}
#endif

