#ifndef DV_NET_SNDRCV_H
#define DV_NET_SNDRCV_H
// $Id: message.h,v 1.9 2003/07/07 11:25:37 dvermeir Exp $
#include <strstream>
#include <string>
#include <iostream>
#include <stdexcept>

namespace Dv {
namespace Net {

/**
 * A SimpleMessage wraps a byte array which it may or may not
 * own. A SimpleMessage can be sent to a stream and it can
 * received from a stream. The protocol ensures that any
 * byte array can be transferred.
 */
class SimpleMessage {
public:
  /**
   * Constructor. 
   * \param data pointer to byte array.
   * \param size of byte array
   * \warning The array will not be copied and the message will
   *   not be its owner.
   */
  SimpleMessage(const char* data = 0, size_t size = 0): 
    data_(data), size_(size), owner_(false) {};
  /**
   * Destructor. Delete the data if we are the owner.
   */
  virtual ~SimpleMessage() {
    if (owner_)
      delete data_;
    }

  /** \return pointer to data char array \warning this is not a C
   * string, see SimpleMessage::size */
  const char* data() const { return data_; }
  /** \return size of data array */
  const size_t size() const { return size_; }

  /**
   * Send data byte-array.
   * The following protocol is used:
   * @code
   * size-of-encoded-byte-array <newline> encoded-byte-array
   * @endcode
   * Test stream status to find out whether the operation succeeded.
   * @param os stream to send to.
   * @return os
   * @sa SimpleMessage::receive
   */
  std::ostream& send(std::ostream& os) const;
  /**
   * Receive data byte-array sent by SimpleMessage::send.
   * @param is stream to receive data from.
   * @return is
   * @sa SimpleMessage::send
   * @exception runtime_error upon any error
   * @warning the message will become the owner of the byte array.
   */
  std::istream& receive(std::istream& is) throw (std::runtime_error);

protected:
  /**
   * Replace underlying byte array.
   * \param p new byte array
   * \param size of new byte array
   * \return p
   * \warning the message will not be the owner of the new byte array.
   */
  const char* data(const char* p, size_t size);
  /**
   * Replace byte array by a newly allocated array of size bytes.
   * Note that the returned pointer is not const, so the derived class
   * can update the byte array.
   * \param size of new byte array
   * \return pointer to newly allocated byte array.
   * \warning The message owns the new byte array, so the derived
   * class should not delete it.
   */
  char* data(size_t size);
private:
  const char*	data_;
  size_t 	size_;
  bool		owner_; // are we owner of data_
};

/**
 * Template extension of SimpleMessage. Using this template, anything
 * that can be printed (using \c operator>>) can be sent and anything
 * that can be read using \c operator<< can be received. 
 * The template parameter type also needs a default constructor.
 */
template<typename T>
class Message: public SimpleMessage {
public:
  /** Constructor. Stores output of <code>operator<<(std::ostream&, const T&)</code>) */
  Message(const T& t = T()) throw (std::runtime_error) {
    if (! (oss_ << t) )
      throw std::runtime_error("Message<T>::Message(const T&) cannot store T object");
    data(oss_.str(), oss_.pcount());
    }
  /** Destructor. */
  ~Message() {
    oss_.freeze(0);
    }
  /** Reconstruct <code>T</code> by reading from stringstream on data */
  operator T() const throw (std::runtime_error) {
    std::istrstream iss(data(), size());
    T	t;
    if ( ! (iss >> t) ) // this will not work for string
      throw std::runtime_error("Message<T>::operator T() cannot read T object");
    return t;
    }
private:
  std::ostrstream oss_;
};

/**
 * Specialization for string.
 * \sa Message<T>
 */
template<>
class Message<std::string>: public SimpleMessage {
public:
  /** Constructor. Simply stores string. */
  Message(const std::string& s) {
    data(s.data(), s.size());
    }
  /** Destructor. */
  ~Message() {}
  /** Reconstruct string. */
  operator std::string() const { return std::string(data(), size()); }
};


}}

inline std::ostream&
operator<<(std::ostream& os, const Dv::Net::SimpleMessage& m) {
return m.send(os);
}

inline std::istream&
operator>>(std::istream& is, Dv::Net::SimpleMessage& m) throw (std::runtime_error) {
return m.receive(is);
}

#endif
