// $Id: serversocket.C,v 1.11 2004/12/22 11:10:21 dvermeir Exp $

#include <config.h>
#include <string.h> // strerror
#include <errno.h>
#include <unistd.h> // close(int)
#include <dvutil/tostring.h>
#include <dvnet/serversocket.h>

#ifdef HAVE_ARPA_INET_H
#include <arpa/inet.h> // sockaddr_in, INADDR_ANY, int htonl(), int htons(), int ntohs()
#endif

#ifdef HAVE_NETINET_IN_H
#include <netinet/in.h> // for *BSD
#endif

#ifdef HAVE_SYS_SOCKET_H
#include <sys/socket.h> // AF_INET, SOCK_STREAM, SOL_SOCKET, SO_REUSEADDR, setsockopt()
#endif

namespace Dv {
namespace Net {

const std::string 
ServerSocketError::NAME("ServerSocketError: ");

//////////////// constructor, destructor 

ServerSocket::ServerSocket(int port, int backlog) throw (ServerSocketError): 
  server_address_(new struct sockaddr_in), closed_(false) {
const static std::string FUNC("ServerSocket::ServerSocket");

// Create socket fd.
if ((socket_fd_=socket(AF_INET,SOCK_STREAM,0))<0) {
  char* sysmsg = strerror(errno);
  throw ServerSocketError(FUNC+"("+Dv::Util::tostring(port)+"): socket() failed\n\t"+sysmsg);
  }

int option = 1;	
if (setsockopt(socket_fd_,SOL_SOCKET,SO_REUSEADDR,
     reinterpret_cast<char*>(&option),sizeof(option))<0) {
  char* sysmsg = strerror(errno);
  throw ServerSocketError(FUNC+"("+Dv::Util::tostring(port)+"): setsockopt() failed\n\t"+sysmsg);
  }

// Bind the server address to the socket.
struct sockaddr_in* sa_in = static_cast<struct sockaddr_in*>(server_address_);
memset(sa_in,0,sizeof(sockaddr_in));
sa_in->sin_family = AF_INET;
sa_in->sin_addr.s_addr = htonl(INADDR_ANY);
sa_in->sin_port	= htons(port);

struct sockaddr* sa = static_cast<struct sockaddr *>(server_address_);
if (bind(socket_fd_,sa, sizeof(sockaddr_in))<0) {
  char* sysmsg = strerror(errno);
  throw ServerSocketError(FUNC+"("+Dv::Util::tostring(port)+"): bind failed\n\t"+sysmsg);
  }

// Get final server_address_ (e.g. port may be different if it was 0)
socklen_t server_address_len = sizeof(struct sockaddr_in);
getsockname(socket_fd_,sa,&server_address_len);

// Listen for connections on socket.
if (listen(socket_fd_,backlog)<0) {
  char* sysmsg = strerror(errno);
  throw ServerSocketError(FUNC+"("+Dv::Util::tostring(port)+"): listen failed\n\t"+sysmsg);
  }
}

ServerSocket::~ServerSocket() {
close();
struct sockaddr_in* sa_in = static_cast<struct sockaddr_in*>(server_address_);
delete sa_in;
}

int
ServerSocket::port() {
struct sockaddr_in* sa_in = static_cast<struct sockaddr_in*>(server_address_);
return ntohs(sa_in->sin_port);
}

unsigned long
ServerSocket::address() {
struct sockaddr_in* sa_in = static_cast<struct sockaddr_in*>(server_address_);
return sa_in->sin_addr.s_addr;
}

Dv::Util::ref<InetAddress>
ServerSocket::inetaddress() {
return InetAddress::by_address(address());
}

int
ServerSocket::fd_accept() {
if (closed_)
  return -1;

struct sockaddr_in  client_address;
socklen_t  client_address_len = sizeof(client_address);
int  client_socket_fd;

do {
  client_socket_fd = ::accept(socket_fd_,
			reinterpret_cast<struct sockaddr *>(&client_address),
			&client_address_len);
   }
while ( (client_socket_fd <0) && (errno == EINTR));

if (client_socket_fd<0) 
  return -1;

return client_socket_fd;
}

bool
ServerSocket::connection(time_t timeout, int* err_ptr) {
int syserr;
if (! err_ptr)
  err_ptr = &syserr;
return Dv::Util::fdstreambuf::fdwait(fd(), timeout, *err_ptr);
}

Dv::Util::ref<Socket>
ServerSocket::accept(time_t delay,size_t bufsz,bool non_blocking, std::ostream* dbg) {
return Socket::buf2socket(
        new Dv::Util::fdstreambuf(fd_accept(),bufsz,bufsz,delay,non_blocking, dbg) );
}

void
ServerSocket::close() {
if (closed_)
  return;
::shutdown(socket_fd_, SHUT_RDWR);
::close(socket_fd_);
closed_ = true;
socket_fd_ = -1;
}

}}
