#ifndef	DV_SERVER_SOCKET_H
#define	DV_SERVER_SOCKET_H

// $Id: serversocket.h,v 1.10 2003/10/31 14:37:19 dvermeir Exp $

#include <dvnet/inetaddress.h>
#include <dvnet/socket.h>
#include <stdexcept>
/** @file
 *  A Dv::Net::ServerSocket object represents an internet server.
 *  Dv::Net::ServerSocket::accept() returns a (reference to) a
 *  Dv::Net::Socket representing a client connection as an iostream.
 */

namespace Dv {
namespace Net {
/** A runtime exception thrown by some ServerSocket operations. */
class ServerSocketError: public std::runtime_error {
public:
  /** Name of class, is prepended to each message argument of the constructor. */
  static const std::string NAME;
  /** Constructor, prepends NAME to message to obtain * std::runtime_error::what(). */
  ServerSocketError(const std::string& message): std::runtime_error(NAME+": "+message) {}
};


/** A class representing an internet server.
 * Example usage:
 * @code
 * try {
 *   Dv::Net::ServerSocket server(server-port);
 *   Dv::Util::ref<Socket> client(server.accept());
 *   std::cerr << "Connection from " << client.host() << ":" << client.port() << std::endl;
 *
 *   std::string request;
 *   std::string reply;
 *
 *   while (*client>>request)
 *     *client << reply;
 *   }
 * catch (Dv::Net::ServerSocketError& e) {
 *   std::cerr << e.what() << std::endl;
 *   return 1;
 *   }
 * @endcode
*/
class ServerSocket {
public:
  /** Start a server on port.
   * @param port on which server will listen for connections.
   * @param backlog no of connections allowed in backlog.
   * @exception Dv::Net::ServerSocketError if anything goes wrong.
   */
  explicit ServerSocket(int port=0,int backlog = 10) throw (ServerSocketError);
  /** Destructor; also closes the socket. */
  ~ServerSocket(); 
  /** @return the port on which server listens. */
  int port();
  /** @return the numeric internet address of the server host */
  unsigned long	 address();
  /** @return the internet address of server.
   * @sa Dv::Net::InetAddress
   */
  Dv::Util::ref<InetAddress> inetaddress();
  /** Accept a new connection from a client.
   * @param delay in milliseconds: the time allowed for any
   *  subsequent I/O operation on the returned Socket. A delay of 0
   *  implies that I/O operations will wait indefinitely.
   * @param bufsize the size of the buffer in the returned Socket.
   * @param non_blocking whether the underlying fdstreambuf for the
   *   new Socket should be non-blocking
   * @param dbg pointer to ostream on which debug info for the new
   *   Socket will be written or 0 for no debug
   * @return A new socket representing a connection with a client.
   *   The status of the returned Socket must be tested to
   *   verify whether the operation was succesful.
   * @see Dv::Util::fdstreambuf
   */
  Dv::Util::ref<Socket> accept(time_t delay=0, size_t bufsize=1024,
    bool non_blocking=false, std::ostream* dbg=0);

  /** Check whether connections are waiting to be accepted.
   * @param timeout in milliseconds:
   * @param syserr if not 0, it will contain a copy of ::errno
   *   upon failure
   * @return true iff accept should succeed, no iff if no input activity on this
   *   serversocket was detected for @a timeout millisecs
   *
   * Example:
   * @code
   * Dv::Net::ServerSocket ss(port);
   * ..
   * if ( ss.connection(2000) ) { // only if a connection is waiting within 2 sec
   *   Dv::Util::ref<Dv::Net::Socket> so(ss.accept());
   *   *so << "+OK" << std::endl;
   *   }
   * else {
   *   // do something else
   *   }
   * @endcode
   * @see Dv::Util::fdstreambuf::fdwait
   */
  bool connection(time_t timeout, int* syserr = 0);

  /** @return underlying file descriptor (-1 if it was closed). */
  int  fd() const  { return (socket_fd_);}
  /** Closes underlying socket, makes ServerSocket unusable by
   * performing shut_down(2) en close(2) on it and setting
   * the underlying file descriptor to -1.
   */
  void close();
protected:
  /** Auxiliary function for accept(). 
   * @return new socket connected with client
   */
  int  fd_accept(); // used by ServerSocket::accept()
private:
  int	socket_fd_;
  void*	server_address_; // a pointer to a sockaddr_in 
  // sever_address_ is opaque so we don't need to include config.h
  bool	closed_;
  ServerSocket(const ServerSocket&); // forbidden
  ServerSocket&	operator=(const ServerSocket&); // forbidden
};

}}
#endif
