// $Id: socket.C,v 1.15 2004/12/27 09:26:04 dvermeir Exp $

#include <config.h>
#include <string.h>

#if HAVE_UNISTD_H
# include <unistd.h> // for close(), sleep()
#else
extern "C" {
extern int close(int fd);
extern int dup(int fd);
extern unsigned int sleep(usigned int secs);
}
#endif

#include <errno.h>

#ifdef HAVE_SYS_SOCKET_H
// for AF_INET, SOCK_STREAM, SOL_SOCKET, SO_REUSEADDR, setsockopt(),
#include <sys/socket.h>
#endif

#ifdef HAVE_ARPA_INET_H
#include <arpa/inet.h>	// for struct sockaddr_in
#endif

#ifdef HAVE_NETINET_IN_H
#include <netinet/in.h> // for *BSD
#endif

#include "socket.h"

using namespace std;
using namespace Dv::Util;

static const char* const MSG_OK = "ok";
static const char* const MSG_ERROR = "error";
static const char* const MSG_CLOSED = "closed";
static const char* const MSG_UNKNOWN_HOST = "unknown host";
static const char* const MSG_ACCEPT_ERROR = "accept error";
static const char* const MSG_WRONG_FAMILY = "peer not AFINET";
static const char* const MSG_EOFBIT = "eof on stream::iostate";
static const char* const MSG_BADBIT = "badbit on stream::iostate";
static const char* const MSG_FAILBIT = "failbit on stream::iostate";

#include <sstream>

static string
socket_debug(const string& member_name,const Dv::Net::Socket* s) {
std::ostringstream oss;
oss << "Socket::" << member_name << "(" << s;
return oss.str();
}

namespace Dv {
namespace Net {

///////////////	constructors, destructor

// Regular client constructor.
Socket::Socket(const string& host,int port,size_t bufsz,time_t delay,
  bool non_blocking, ostream* dbg):
	iostream(new fdstreambuf(mkfd(),bufsz,bufsz,delay,non_blocking)), 
        host_(host), port_(port), errno_(OK), connected_(false),
	debug_(dbg) {
if (debug())
  *debug_ << socket_debug("Socket",this) 
  	  << ", host=" << host 
	  << ", port=" << port 
  	  << ", bufsz=" <<  bufsz 
	  << ", delay=" << delay 
	  << ", non_blocking=" << non_blocking 
	  << ", debug=" << dbg 
	  << ")" << endl;
connect();
}

// Copy ctor.
Socket::Socket(const Socket& s): 
  iostream(new fdstreambuf(::dup(s.sfd()), s.rdbuf()->inbuf_sz(), s.rdbuf()->outbuf_sz(),
           s.rdbuf()->delay(), s.rdbuf()->non_blocking())),
  host_(s.host()), port_(s.port()), errno_(s.errno_),
  connected_(s.connected()), debug_(s.debug_) {
if (debug())
  *debug_ << socket_debug("Socket[cctor]",this) 
  	  << ", host=" << host() 
	  << ", port=" << port()
  	  << ", inbufsz=" <<  rdbuf()->inbuf_sz() 
  	  << ", outbufsz=" <<  rdbuf()->outbuf_sz() 
	  << ", delay=" << rdbuf()->delay() 
	  << ", non_blocking=" << rdbuf()->non_blocking() 
	  << ", debug=" << debug()
	  << ")" << endl;
}

ref<Socket>
Socket::dup() {
return ref<Socket>(new Socket(*this));
}

// Special protected constructor for derived classes, note that fdstreambuf
//   must have been dynamically allocated using new(), the destructor will delete it.
Socket::Socket(const string& host, int port, fdstreambuf* buf,
  ostream* dbg): iostream(buf), host_(host), port_(port), errno_(OK), 
    connected_(false), debug_(dbg) {
if (debug())
  *debug_ << socket_debug("Socket",this)
   	  << ", host=" << host 
	  << ", port=" << port 
  	  << ", fdstreambuf=" << buf 
	  << ")" << endl;
connect();
}

// Special private constructor for use by ServerSocket (through Socket::fd2socket()).
Socket::Socket(fdstreambuf* buf, ostream* dbg): iostream(buf), 
  host_(), port_(0), errno_(OK), connected_(true), debug_(dbg) {
if (debug())
  *debug_ << socket_debug("Socket",this)
  	  << ", fdstreambuf=" << buf 
	  << ")" << endl;
if (sfd()<0) {
  error(ACCEPT_ERROR);
  connected_ = false;
  return;
  }
if (!get_peer())  // needed to fill in host, port
  connected_ = false;
}

Socket::~Socket() {
if (debug())
  *debug_ << socket_debug("~Socket",this) << ")" << endl;
if (rdbuf())
  delete rdbuf(); // this does all the work
iostream::rdbuf(0);
}

///////////////	mkfd make socket (and associated fd) return fd or -1 upon error
int
Socket::mkfd() {
int syserr;
int fd(mksocketfd(syserr));
if (fd<0)
  error(syserr);
return fd;
}

///////////////	static mkfd makes socket (and associated fd) return fd or -1 upon error
int
Socket::mksocketfd(int& syserr) {
int fd;
int option(1);	

if ((fd = socket(AF_INET, SOCK_STREAM , 0))<0) {
  syserr = errno;
  return -1;
  }
if (setsockopt(fd,SOL_SOCKET,SO_REUSEADDR,
      reinterpret_cast<char*>(&option),sizeof(option))<0) {
  syserr = errno;
  return -1;
  }
return fd;
}

///////////////	connect & disconnect
bool
Socket::connect(unsigned int delay) { // return false if did not succeed
if (debug()) 
  *debug_ << socket_debug("connect",this)
  	  << ", delay=" << delay 
          << ", connected()=" << ( connected() ? "true" : "false" )
          << ", sfd()=" << sfd() 
	  << ")" << endl;

ostream* sbuf_debug(rdbuf() ? rdbuf()->debug() : 0);

if (connected())
  return true;

ref<InetAddress> ia = InetAddress::by_name(host_);
if (!ia) {
  error(UNKNOWN_HOST);
  return false;
  }

int fd = sfd();
if (fd<0)  { // e.g. because of earlier connect() failure
  rdbuf()->fd(mkfd());
  fd = sfd();
  }

struct sockaddr_in server_address;
memset(&server_address,0,sizeof(server_address));
server_address.sin_family	= AF_INET;
server_address.sin_addr.s_addr	= ia->address();
server_address.sin_port		= htons(port_);

int	status;

// Upon successful completion, connect() returns 0.
// Otherwise, -1 is returned and errno is set to indicate the error.

do {
  status = ::connect(fd,reinterpret_cast<struct sockaddr *>(&server_address), 
                     sizeof(server_address));
  if (status<0) {
    if (errno == EINPROGRESS) { // non-blocking I/O
      { /// debug
      if (debug()) 
        *debug_ << socket_debug("connect",this) << "): "
		<< "errno = " << EINPROGRESS 
	        << " (EINPROGRESS)" << endl;
      if (sbuf_debug && ( sbuf_debug != debug() ) )
        *sbuf_debug << socket_debug("connect",this) << "): "
		    << "errno = " << EINPROGRESS 
	            << " (EINPROGRESS)" << endl;
      }
      int syserrno;
      // wait for output
      if (fdstreambuf::fdwait(fd, delay, syserrno, true) > 0) {
	// then check socket status
        socklen_t status_size(sizeof(status));
        getsockopt(fd, SOL_SOCKET, SO_ERROR, &status, &status_size);
	{ /// debug
	if (debug())
          *debug_ << socket_debug("connect",this) << "): "
	  	 << "getsockopt status => " << status << endl;
        if (sbuf_debug && ( sbuf_debug != debug() ) )
          *sbuf_debug << socket_debug("connect",this) << "): "
	  	 << "getsockopt status => " << status << endl;
	}
        if (status==0)
          error(OK);
        else
          error(status);
        }
      else {
	if (debug())
          *debug_ << socket_debug("connect",this) << "): " 
	  	  << "fdwait erro => " << syserrno << endl;
        if (sbuf_debug && ( sbuf_debug != debug() ) )
          *sbuf_debug << socket_debug("connect",this) << "): " 
	  	      << "fdwait erro => " << syserrno << endl;
        error(syserrno);
        }
      }
    else  
      error(errno);
    }
  else
    error(OK);
  }
while ((status==-1)&&(error()==EINTR));
connected_ = (status == 0);

if (debug()) 
  *debug_ << socket_debug("connect",this) << "): "
  	 << "status = " << status << endl;
if (sbuf_debug && ( sbuf_debug != debug() ) )
  *sbuf_debug << socket_debug("connect",this) << "): "
  	 << "status = " << status << endl;

if (connected()) 
  error(OK);
else
  rdbuf()->close();
return connected();
}

void
Socket::close() {
if (debug()) 
  *debug_ << socket_debug("close",this) << ")" << endl;
::shutdown(rdbuf()->fd(), SHUT_RDWR);
rdbuf()->close();
error(CLOSED);
connected_ = false;
}

///////////////	get_peer fills in host_, port_ from connection on sfd()

bool
Socket::get_peer() {
if (debug()) 
  *debug_ << socket_debug("get_peer",this) << ")" << endl;
struct sockaddr_in	peer_address;
socklen_t		namelen	= sizeof(peer_address);

if (getpeername(sfd(),reinterpret_cast<struct sockaddr *>(&peer_address),
     &namelen)<0) {
  error(errno);
  return false;
  }

if (peer_address.sin_family!= AF_INET) {
  error(WRONG_FAMILY);
  return false;
  }

ref<InetAddress> ia = InetAddress::by_address(peer_address.sin_addr.s_addr);
if (ia)
  host_	= ia->host();
else	// use dot notation
  host_	= inet_ntoa(peer_address.sin_addr);
port_	= ntohs(peer_address.sin_port);
return true;
}

bool
Socket::timedout() const {
// The first component of the disjunction checks the connect() status.
return ((error()==ETIMEDOUT)||rdbuf()->timedout()); 
}

bool
Socket::timedout(bool new_timedout_status) {
// The first component of the disjunction checks the connect() status.
if (new_timedout_status == false) {
  if (timedout()) {
    error(0);
    if (rdbuf())
      rdbuf()->timedout(false);
    return true;
    }
  else
    return false;
  }
else {
  if (error() == 0) {
    error(ETIMEDOUT);
    if (rdbuf())
      rdbuf()->timedout(true);
    return true;
    }
  else
    return false;
  }
}

ref<Socket>
Socket::buf2socket(fdstreambuf* buf) {
return ref<Socket>(new Socket(buf));
}

///////////////	error recording

void
Socket::error(int no) {
if (debug()) 
  *debug_ << socket_debug("error",this) 
	  << ", no=" << no
          << ")" << endl;
errno_ = no;
if (no)
  setstate(ios::failbit);
else {
  clear(); // setstate does not clear possible badbit
  setstate(ios::goodbit);
  }
}

int
Socket::error() const {
if (errno_)
  return errno_;
if (rdbuf()) {
  int rdbuf_err(rdbuf()->error());
  if (rdbuf_err)
    return rdbuf_err;
  }
if (eof())
  return EOFBIT;
if (fail())
  return FAILBIT;
if (bad())
  return BADBIT;
return 0;
}

string
Socket::strerror() const {
switch (error()) {
  case OK:	     return MSG_OK;
  case CLOSED:       return MSG_CLOSED;
  case UNKNOWN_HOST: return MSG_UNKNOWN_HOST;
  case ACCEPT_ERROR: return MSG_ACCEPT_ERROR;
  case WRONG_FAMILY: return MSG_WRONG_FAMILY;
  case EOFBIT: 	     return MSG_EOFBIT;
  case FAILBIT:	     return MSG_FAILBIT;
  case BADBIT:	     return MSG_BADBIT;
  default: 
    if (error()>0) {
      char* msg = ::strerror(error());
      return string(msg);
      }
    else
      return MSG_ERROR;
  }
} 

ostream*
Socket::debug(ostream* dbg) {
ostream* tmp(debug_);
debug_ = dbg;
return tmp;
}

ostream*
Socket::debug() {
return debug_;
}

}}
