#ifndef	DV_SOCKET_H
#define	DV_SOCKET_H
// $Id: socket.h,v 1.15 2003/03/04 18:03:02 dvermeir Exp $

#include	<iostream>
#include	<dvutil/ref.h>
#include	<dvutil/fdstreambuf.h>
#include	<dvnet/inetaddress.h>
/** \file
 *  The Dv::Net::Socket class provides a network connection as an iostream.
 */
namespace Dv {
namespace Net {
/** 
 * Provide a network connection as an iostream.
 *
 * Example usage:
 *   \code
 *   Socket	so("tinf2.vub.ac.be",8000)
 *   if (so) {
 *     so << "request" << endl;
 *     string line;
 *     while (getline(so,line))
 *       process_answer(line);
 *     if (so.timedout())
 *       cerr << "Connection timed out";
 *     }
 *   else {
 *     cerr << so.strerror() << endl;
 *     }
 *   \endcode
 */
class Socket: public std::iostream {
public:
  /**
   * Status codes, positive numbers are reserved for copies of the system errno.
   */
  enum SOCKET_ERRORS { 
  	OK = 0, UNKNOWN_HOST = -1, CLOSED = -2, 
  	 ACCEPT_ERROR = -3, WRONG_FAMILY = -4, 
	 EOFBIT = -5, BADBIT = -6, FAILBIT = -7,
	 MAX_ERRNO = -7 };
  /** 
   * Set up a client connection to a  host:port.
   *
   * \param host name of host to connect to.
   * \param port number of port to connect to.
   * \param bufsz size (in bytes) of input and output buffers.
   * \param delay (in milliseconds) time allowed for any I/O operation
   *   to complete. A value of 0 means ``wait forever''. Note
   *   that timedout() makes no sense if delay is 0.
   * \param non_blocking if true, the underlying fdstreambuf will be
   *   non-blocking which prevents some cases of blocking I/O operations.
   * \param dbg pointer to stream to write debug to or 0
   *
   * \see timedout, connected, Dv::Util::fdstreambuf, debug
   */
  Socket(const std::string& host, int port, size_t bufsz=1024,
    time_t delay=0, bool non_blocking=false, std::ostream* dbg=0);

  /** Copy ctor.  This function constructs a new Socket (iostream) and
   * associated fdstreambuf, based a new filedescriptor, obtained from
   * this Socket's file descriptor This function was introduced to
   * allow concurrent reading and writing to a connection (from
   * different threads). Experiments to do this failed when using the
   * same Socket object. If the writer uses a copy of the original
   * (which is used by the reader), the experiment works fine.
   * \param so socket whose file descriptor will be used to construct a
   *   ``copy'' 
   * \warning Closing a socket also shuts it down using shutdown(2) with
   * the SHUT_RDWR option, making all copies of this Socket unusable. However,
   * the underlying file descriptors of these copies are not closed. To
   * avoid a memory leak, they should be closed as well.
   */

  Socket(const Socket& so);
  /** 
   * See the copy constructor.
   * \return reference to copy of this Socket
   * \see Socket::Socket(const Socket& so)
   */
  Dv::Util::ref<Socket> dup();
  
  /**
   * Also closes socket using close().
   *
   * \see close()
   */
  ~Socket();
  /**
   * Close underlying socket. This also does a \a shutdown(2) with the
   * \a SHUT_RDWR which prevents any transmissions and should
   * make all pending I/O calls return with an error.
   *
   * \see ~Socket
   */
  void	close();
  /** 
   * Try to (re)connect to the same host/port.
   *
   * \param delay (in milliseconds) time allowed for the connection to complete.
   *   A value of 0 means ``wait forever''.
   * \return true iff the connect operation succeeded.
   *
   * If the constructor failed, it is possible to retry using connect().
   * This may occur e.g. if the client is able to launch the server.
   * Note the delay parameter: when launching the server, at least
   * Solaris 7 needs a delay (e.g. 3 seems to work find).
   */
  bool connect(unsigned int delay=0);
  /**
   * \return true iff last operation timed out 
   * \see Socket::Socket()
   */
  bool timedout() const;
  /**
   * Set timedout flag. Usually, the argument will be "false", such
   * that further I/O on the socket becomes possible.
   * @param new_timedout_status: true or (usually) false
   * @return true iff status change succeeded. It will fail e.g. if
   *  the argument is "false" and the socket had not actually timed out,
   *  or if the argument is "true" and socket status was not Socket::OK.
   * \code
   * ref<Socket> s;
   * size_t try(0);
   *
   * while ( (! getline(*s, line) ) && (++try <3) )
   *   if (s->timedout()) {
   *     s->timedout(false);
   *   else {
   *	 std::cerr << "I/O error: " << s->strerror() << std::endl;
   *     break;
   *     } 
   * \endcode
   * \see Socket::Socket() Socket::timedout()
   */
  bool timedout(bool new_timedout_status);
  /**
   * \return port number of Socket.
   */
  int port() const 	{ return port_; }
  /**
   * \return host name of Socket.
   */
  const std::string&	host() const	{ return host_; }
  /**
   * \return true iff last call to connect() was succesful.
   * \see connect, Socket::Socket
   */
  bool connected() const { return connected_; }
  /**
   * \return status of Socket, only 0 is ok.
   * 
   * A status of 0 means ok, positive integers correspond to system
   * errno values (man 2 errno). Negative numbers correspond to
   * ``specific'' errors for the Socket class (or its subclasses).
   *
   * The values EOFBIT, FAILBIT or BADBIT will be returned if the
   * corresponding bit in iostate is turned on and no other specific
   * error was detected.
   *
   * \see SOCKET_ERRORS
   */
  int error() const;
  /**
   * Return string representation of error().
   * This function is virtual because subclasses may want to add
   * their own error messages (but the convention of using negative
   * numbers for local error codes and positive ones for ::errno
   * values should be adhered to).
   */
  virtual std::string strerror() const;
  /**
   * \return fdstreambuf associated with this socket or 0 if none.
   */
  Dv::Util::fdstreambuf* rdbuf() const { 
    return static_cast<Dv::Util::fdstreambuf*>(std::iostream::rdbuf()); }
  /**
   * \return underlying file descriptor (may be <0 if no connection).
   */
  int sfd() const      { return (rdbuf()->fd()); }
  /**
   * Create a new socket descriptor.
   *
   * \param syserr will contain system errorcode (::errno) if the
   *   socket could not be created (and the return value is -1).
   * \return descriptor of a new socket or -1.
   * \see socket(2)
   */
  static int mksocketfd(int& syserr);
  /**
   * Set a stream to write debug information to. A parameter 0 is
   * allowed and will turn off debugging. Note that this debug status
   * is independent of the Socket::rdbuf()->debug() facility which
   * can be used independently.
   *
   * \param dbg pointer to stream to write debug to or 0
   * \return previous value of debug()
   */
  std::ostream* debug(std::ostream* dbg);
  /**
   * \return pointer to current debug stream or 0.
   */
  std::ostream* debug();
protected:
  /**
   * Create a new socket and return associated fd, or -1 upon failure.
   * Upon failure, error will be set to the systems's errno.
   *
   * \return descriptor of a new socket or -1.
   * \see mksocketfd, error
   */
  int mkfd();
  /**
   * Set error status.
   *
   * \param e new error status.
   * \see error()
   */
  void error(int e);

  /**
   * Constructor for use by derived classes.
   * This constructor can e.g. be used by a Dv::Socket specialization
   * that encrypts the data traffic (by means of a class derived from
   * fdstreambuf). This function does not attempt to connect to
   * the given host/port.
   *
   * \param host name of host to connect to.
   * \param port number of port to connect to.
   * \param buffer fdstreambuf to use. This buffer must have been dynamically
   *   allocated because Socket::~Socket() will delete it.
   * \param dbg pointer to stream to write debug to or 0
   * \see Socket::Socket, debug
   */
  Socket(const std::string& host, int port, Dv::Util::fdstreambuf* buffer, std::ostream* dbg = 0);
  /**
   * Constructor for use by derived classes.
   * This constructor can e.g. be used by a Dv::Socket specialization
   * that encrypts the data traffic (by means of a class derived from
   * fdstreambuf). Socket::get_peer() will be used to fill in host and port.
   *
   * \param buffer fdstreambuf to use. This buffer must have been dynamically
   *   allocated because Socket::~Socket() will delete it.
   * \param dbg pointer to stream to write debug to or 0
   * \see get_peer, debug
   */
  Socket(Dv::Util::fdstreambuf* buffer, std::ostream* dbg = 0);
private:
  /**
   * Host this socket refers to.
   */
  std::string			host_;
  /**
   * Port number this socket refers to.
   */
  int				port_;
  /**
   * Error status of socket.
   * \see error()
   */
  mutable int			errno_;
  /**
   * Whether socket is connected.
   * \see connected()
   */
  bool				connected_;
  /**
   * Pointer to debug output stream.
   * \see debug()
   */
  std::ostream*			debug_;

  /**
   * No copy ctor.
  Socket(const Socket&); // forbidden
   */
  /**
   * No assignment.
   */
  Socket& operator=(const Socket&); // forbidden

  /**
   * Set host_ and port_.
   * \see getpeername(2).
   */
  bool get_peer();

  /**
   * Friend class that can use buf2socket()
   */
  friend class ServerSocket;	
  /**
   * Used by ServerSocket.
   */
  static Dv::Util::ref<Socket> buf2socket(Dv::Util::fdstreambuf*);
};

}}

#endif

